package http

import (
	"bufio"
	"compress/gzip"
	"context"
	"errors"
	"io"
	"net"
	"net/http"
	"sync"

	"git.nobla.cn/golang/aeus/pkg/logger"
	"github.com/gin-gonic/gin"
)

const (
	Protocol = "http"
)

type (
	Option func(*options)

	options struct {
		network    string
		address    string
		certFile   string
		keyFile    string
		debug      bool
		hander     http.Handler
		logger     logger.Logger
		context    context.Context
		ginOptions []gin.OptionFunc
		enableCORS bool
	}

	HandleFunc func(ctx *Context) (err error)

	Middleware func(http.Handler) http.Handler

	httpMetadataReader struct {
		hd http.Header
	}

	httpMetadataWriter struct {
		w http.ResponseWriter
	}
)

const (
	headerAcceptEncoding  = "Accept-Encoding"
	headerContentEncoding = "Content-Encoding"
	headerVary            = "Vary"
)

var (
	gzPool           sync.Pool
	assetsExtensions = []string{".css", ".js", ".png", ".jpg", ".jpeg", ".gif", ".svg", ".ico", ".woff", ".woff2", ".ttf", ".eot", ".otf"}
)

type gzipWriter struct {
	gin.ResponseWriter
	writer *gzip.Writer
}

func (g *gzipWriter) WriteString(s string) (int, error) {
	g.Header().Del("Content-Length")
	return g.writer.Write([]byte(s))
}

func (g *gzipWriter) Write(data []byte) (int, error) {
	g.Header().Del("Content-Length")
	return g.writer.Write(data)
}

func (g *gzipWriter) Flush() {
	_ = g.writer.Flush()
	g.ResponseWriter.Flush()
}

// Fix: https://github.com/mholt/caddy/issues/38
func (g *gzipWriter) WriteHeader(code int) {
	g.Header().Del("Content-Length")
	g.ResponseWriter.WriteHeader(code)
}

var _ http.Hijacker = (*gzipWriter)(nil)

// Hijack allows the caller to take over the connection from the HTTP server.
// After a call to Hijack, the HTTP server library will not do anything else with the connection.
// It becomes the caller's responsibility to manage and close the connection.
//
// It returns the underlying net.Conn, a buffered reader/writer for the connection, and an error
// if the ResponseWriter does not support the Hijacker interface.
func (g *gzipWriter) Hijack() (net.Conn, *bufio.ReadWriter, error) {
	hijacker, ok := g.ResponseWriter.(http.Hijacker)
	if !ok {
		return nil, nil, errors.New("the ResponseWriter doesn't support the Hijacker interface")
	}
	return hijacker.Hijack()
}

func newGzipWriter() (writer *gzip.Writer) {
	v := gzPool.Get()
	if v == nil {
		writer, _ = gzip.NewWriterLevel(io.Discard, gzip.DefaultCompression)
	} else {
		if w, ok := v.(*gzip.Writer); ok {
			return w
		} else {
			writer, _ = gzip.NewWriterLevel(io.Discard, gzip.DefaultCompression)
		}
	}
	return
}

func putGzipWriter(writer *gzip.Writer) {
	gzPool.Put(writer)
}

func WithNetwork(network string) Option {
	return func(o *options) {
		o.network = network
	}
}

func WithCORS() Option {
	return func(o *options) {
		o.enableCORS = true
	}
}

func WithAddress(address string) Option {
	return func(o *options) {
		o.address = address
	}
}

func WithCertFile(certFile string) Option {
	return func(o *options) {
		o.certFile = certFile
	}
}

func WithKeyFile(keyFile string) Option {
	return func(o *options) {
		o.keyFile = keyFile
	}
}

func WithLogger(lg logger.Logger) Option {
	return func(o *options) {
		o.logger = lg
	}
}

func WithContext(ctx context.Context) Option {
	return func(o *options) {
		o.context = ctx
	}
}

func WithDebug(debug bool) Option {
	return func(o *options) {
		o.debug = debug
	}
}

func WithHandler(h http.Handler) Option {
	return func(o *options) {
		o.hander = h
	}
}

func WithGinOptions(opts ...gin.OptionFunc) Option {
	return func(o *options) {
		o.ginOptions = opts
	}
}

func (m *httpMetadataReader) Get(key string) string {
	if m.hd == nil {
		return ""
	}
	return m.hd.Get(key)
}

func (m *httpMetadataWriter) Set(key string, value string) {
	if m.w == nil {
		return
	}
	m.w.Header().Set(key, value)
}
